# This code is part of qmatchatea.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""
The :py:class:`QCEmulator` class enables full-python simulations.

Functions and classes
~~~~~~~~~~~~~~~~~~~~~

"""

# pylint: disable=protected-access, bare-except

import os
import time
import warnings
from copy import deepcopy

import numpy as np
import psutil
import qredtea as qrt
import qtealeaves.tensors as qtt
from qiskit import QuantumCircuit
from qtealeaves.abstracttns.abstract_tn import _AbstractTN
from qtealeaves.convergence_parameters import TNConvergenceParameters
from qtealeaves.emulator import MPIMPS, MPS, TTN
from qtealeaves.mpos import DenseMPO
from qtealeaves.observables import TNObservables
from qtealeaves.simulation.tn_simulation import run_tn_measurements

from .circuit import Qcircuit
from .circuit.observables import QCObservableStep
from .utils import QCBackend, SimulationResults
from .utils.tn_utils import QCOperators
from .utils.utils import QCCheckpoints, QCConvergenceParameters

__all__ = ["QCEmulator", "run_py_simulation"]


class QCEmulator:
    """
    Emulator class to run quantum circuits, powered by either
    TTNs or MPS.


    Parameters
    ----------

    num_sites: int
        Number of sites
    convergence_parameters: :py:class:`QCConvergenceParameters`
        Class for handling convergence parameters. In particular, in the MPS simulator we are
        interested in:
        - the *maximum bond dimension* :math:`\\chi`;
        - the *cut ratio* :math:`\\epsilon` after which the singular values are neglected, i.e.
        if :math:`\\lambda_1` is the bigger singular values then after an SVD we neglect all the
        singular values such that :math:`\\frac{\\lambda_i}{\\lambda_1}\\leq\\epsilon`
    local_dim: int, optional
        Local dimension of the degrees of freedom. Default to 2.
    tensor_backend: TensorBackend, optional
        Contains all the information on the tensors, such as dtype and device.
        Default to TensorBackend() (dtype=np.complex128, device="cpu").
    qc_backend: QCBackend, optional
        Backend for the qmatchatea emulation, containing the backend and other important infos.
        Default to QCBackend() (ansatz="MPS", precision="A", device="cpu")
    initialize: str, optional
        Initialization procedure.
        Default to "vacuum", the 0000...0 state.
        Available: "random", "vacuum", path_to_file
    """

    ansatzes = {"MPS": MPS, "TTN": TTN, "MPIMPS": MPIMPS}

    # pylint: disable-next=too-many-arguments
    def __init__(
        self,
        num_sites,
        convergence_parameters=QCConvergenceParameters(),
        local_dim=2,
        tensor_backend=qtt.TensorBackend(),
        qc_backend=QCBackend(),
        initialize="vacuum",
    ):
        if not isinstance(convergence_parameters, TNConvergenceParameters):
            raise TypeError(
                "convergence_parameters must be of the QCConvergenceParameters class"
            )
        if qc_backend.device != tensor_backend.device:
            raise ValueError(
                "Tensor backend and QCBackend have different devices, "
                + f"{tensor_backend.device} and {qc_backend.device} respectively."
            )

        self._trunc_tracking_mode = convergence_parameters.trunc_tracking_mode
        self._qc_backend = qc_backend

        # Classical registers to hold qiskit informations
        self.cl_regs = {}

        # Observables measured
        self.is_measured = [
            True,
            False,
            False,
            True,
            True,
            True,
            True,
            True,
            True,
            False,
            False,
        ]

        # If a TTN, pad with empty sites until you get to a power of 2 sites
        if self.ansatz == "TTN":
            if num_sites & (num_sites - 1) == 0:
                exponent = np.ceil(np.log2(num_sites))
                num_sites = int(2**exponent)

        # Initialize based on the intialized keyword
        if os.path.isfile(initialize):
            if initialize.endswith(
                "pkl" + self.ansatzes[qc_backend.ansatz.upper()].extension
            ):
                self.emulator = self.ansatzes[qc_backend.ansatz.upper()].read_pickle(
                    filename=initialize
                )
            elif initialize.endswith(
                self.ansatzes[qc_backend.ansatz.upper()].extension
            ):
                self.emulator = self.ansatzes[qc_backend.ansatz.upper()].read(
                    filename=initialize,
                    tensor_backend=tensor_backend,
                    cmplx=np.iscomplex(np.empty(1, dtype=tensor_backend.dtype))[0],
                    order="F",
                )
            else:
                raise IOError(f"Extension {initialize} not supported by QCEmulator")

            self.emulator._tensor_backend = tensor_backend
            self.emulator._convergence_parameters = convergence_parameters
        else:
            self.emulator = self.ansatzes[qc_backend.ansatz.upper()](
                num_sites=num_sites,
                convergence_parameters=convergence_parameters,
                local_dim=local_dim,
                initialize=initialize,
                tensor_backend=tensor_backend,
            )

    @property
    def tensor_backend(self):
        """Tensor backend of the simulation"""
        return self.emulator._tensor_backend

    @property
    def ansatz(self):
        """Ansatz of the emulator"""
        return self._qc_backend.ansatz

    def __getattr__(self, __name: str):
        """
        Check for the attribute in emulator, i.e. the QCEmulator inherits all
        the emulator calls.
        This call is for convenience and for retrocompatibility

        .. warning::
            The method `__getattr__` is called when `__getattribute__` fails,
            so it already covers the possibility of the attribute being in the
            base class
        """
        return self.emulator.__getattribute__(__name)

    @classmethod
    def from_emulator(
        cls, emulator, conv_params=None, tensor_backend=None, qc_backend=QCBackend()
    ):
        """
        Initialize the QCEmulator class starting from an emulator class, i.e. either
        MPS or TTN

        Parameters
        ----------
        emulator : :class:`_AbstractTN`
            Either an MPS or TTN emulator
        conv_params : :class:`TNConvergenceParameters`, optional
            Convergence parameters. If None, the convergence parameters of the emulator
            are used
        tensor_backend: TensorBackend, optional
            Contains all the information on the tensors, such as dtype and device.
            Default to TensorBackend() (dtype=np.complex128, device="cpu").
        qc_backend : QCBackend(), optional
            Backend of the qmatchatea simulation

        Return
        ------
        QCEmulator
            The quantum circuit emulator class
        """
        if not isinstance(emulator, _AbstractTN):
            raise TypeError("The emulator should be a TN emulator class")
        if conv_params is None:
            conv_params = emulator._convergence_parameters
        if tensor_backend is None:
            tensor_backend = emulator._tensor_backend

        simulator = cls(
            emulator.num_sites,
            conv_params,
            emulator.local_dim,
            tensor_backend=tensor_backend,
            qc_backend=qc_backend,
        )
        emulator._convergence_parameters = conv_params
        emulator._tensor_backend = tensor_backend
        simulator.emulator = emulator
        simulator.emulator.convert(
            device=tensor_backend.device, dtype=tensor_backend.dtype
        )

        return simulator

    @classmethod
    def from_tensor_list(
        cls, tensor_list, conv_params=None, tensor_backend=None, qc_backend=QCBackend()
    ):
        """
        Initialize the QCEmulator class starting from a tensor list, i.e. either
        MPS or TTN

        Parameters
        ----------
        tensor_list : list of tensors
            Either an MPS or TTN list of tensors
        conv_params : :class:`TNConvergenceParameters`, optional
            Convergence parameters. If None, the convergence parameters of the emulator
            are used
        tensor_backend: TensorBackend, optional
            Contains all the information on the tensors, such as dtype and device.
            Default to TensorBackend() (dtype=np.complex128, device="cpu").
        qc_backend : QCBackend(), optional
            Backend of the qmatchatea simulation

        Return
        ------
        QCEmulator
            The quantum circuit emulator class
        """
        # A list of lists is a TTN, while a list of tensors is an MPS
        initial_state = cls.ansatzes[qc_backend.ansatz].from_tensor_list(
            tensor_list, conv_params=conv_params, tensor_backend=tensor_backend
        )

        simulator = cls.from_emulator(
            initial_state,
            conv_params=conv_params,
            tensor_backend=tensor_backend,
            qc_backend=qc_backend,
        )

        return simulator

    def meas_projective(
        self, nmeas=1024, qiskit_convention=True, seed=None, unitary_setup=None
    ):
        """See the parent method"""
        return self.emulator.meas_projective(
            nmeas=nmeas,
            qiskit_convention=qiskit_convention,
            seed=seed,
            unitary_setup=unitary_setup,
        )

    def to_statevector(self, qiskit_order=True, max_qubit_equivalent=20):
        """See the parent method"""
        return self.emulator.to_statevector(qiskit_order, max_qubit_equivalent)

    def apply_two_site_gate(self, operator, control, target):
        """Apply a two-site gate, regardless of the position on the chain

        Parameters
        ----------
        operator : QTeaTensor
            Gate to be applied
        control : int
            control qubit index
        target : int
            target qubit index

        Returns
        -------
        singvals_cut
            singular values cut in the process
        """
        local_dim = self.local_dim[0]
        if operator.shape == (local_dim**2, local_dim**2):
            operator = operator.reshape([local_dim] * 4)
        # Reorder for qiskit convention on the two-qubits gates
        if control < target or self.ansatz == "TTN":
            operator = operator.transpose([1, 0, 3, 2])

        singvals_cut = self.apply_two_site_operator(operator, [control, target])

        # Trunc tracking mode is stored in self.emulator._convergence_parameters
        singvals_cut = self.emulator._postprocess_singvals_cut(singvals_cut)

        # Bring to CPU/host if attribute available via some example tensor; must be
        # iso center in case of mixed device
        if isinstance(self.emulator, MPIMPS):
            # Will have problems with mixed-device MPI-MPS, but we can live
            # with this for now. Overwriting `get_tensor_of_site` in MPIMPS
            # is definitely necessary
            tensor = self.emulator[0]
        else:
            # in MPS, iso moves to the right and stays on device, TTN is less
            # obvious
            idx = max([control, target])
            tensor = self.emulator.get_tensor_of_site(idx)

        singvals_cut = tensor.get_of(singvals_cut)

        return [singvals_cut]

    def apply_multi_site_gate(self, operator, sites):
        """
        Apply a n-sites gate, regardless of the position on the chain

        Parameters
        ----------
        operator : QTeaTensor | List[QTeaTensor]
            If a single QTeaTensor, it is the unitary matrix of the
            n-qubits gate. If a List[QTeaTensor] it is already
            written in the MPO form
        sites : List[int]
            Sites to which the operator should be applied

        Returns
        -------
        singvals_cut
            singular values cut in the process
        """
        # This site order could be reversed for the qiskit convention
        site_order = np.argsort(sites)
        local_dim = self.local_dim[sites[0]]
        if isinstance(operator, self.tensor_backend.tensor_cls):
            operator = operator.reshape([local_dim] * len(sites) * 2)
            transpose_idxs = np.arange(operator.ndim).reshape(2, -1)
            transpose_idxs[0, :] = transpose_idxs[0, site_order]
            transpose_idxs[1, :] = transpose_idxs[1, site_order]
            operator.transpose_update(transpose_idxs.reshape(-1))
            operator = DenseMPO.from_matrix(
                operator, sites, local_dim, self._convergence_parameters
            )

        singvals_cut = self.apply_mpo(operator)

        # Avoid errors due to no singv cut
        singvals_cut = np.append(singvals_cut, 0)
        if self._trunc_tracking_mode == "M":
            singvals_cut = max(0, singvals_cut.max())
        elif self._trunc_tracking_mode == "C":
            singvals_cut = (singvals_cut**2).sum()

        if hasattr(singvals_cut, "get"):
            singvals_cut = singvals_cut.get()

        return [singvals_cut]

    def meas_observables(self, observables, operators):
        """Measure all the observables

        Parameters
        ----------
        observables : :py:class:`TNObservables`
            All the observables to be measured
        oeprators : :py:class:`TNOperators`
            List of operators that form the circuit stored in THE CORRECT DEVICE.
            If you are running on GPU the operators should be on the GPU.

        Returns
        -------
        TNObservables
            Observables with the results in results_buffer
        """
        if not isinstance(observables, TNObservables):
            raise TypeError("observables must be TNObservables")

        with warnings.catch_warnings():
            # We use a function that raises a warning for a specific thing we are not interested in.
            # So we filter it out.
            warnings.filterwarnings(
                "ignore",
                message="Tried to compute energy with no effective operators. Returning nan",
            )
            # At the moment, observables are only measured serially
            if self.ansatz == "MPIMPS":
                if self._qc_backend.mpi_settings[-1] < 0:
                    self.emulator.reinstall_isometry_serial()
                else:
                    self.emulator.reinstall_isometry_parallel(
                        self._qc_backend.mpi_settings[-1]
                    )
                rank = self.emulator.rank
                tensor_list = self.emulator.mpi_gather_tn()
                if rank != 0:
                    return observables
                emulator = MPS.from_tensor_list(
                    tensor_list,
                    self.emulator._convergence_parameters,
                    self.tensor_backend,
                )
            else:
                rank = 0
                emulator = self.emulator

            if rank == 0:
                emulator.normalize()
                observables = run_tn_measurements(
                    state=emulator,
                    observables=observables,
                    operators=operators,
                    params={},
                    tensor_backend=self.tensor_backend,
                    tn_type=6 if self.ansatz in ("MPS", "MPSMPI") else 5,
                )

        return observables

    def run_circuit_from_instruction(self, op_list, instr_list):
        """
        Run a circuit from the istructions.

        Parameters
        ----------
        op_list : list of tensors
            List of operators that form the circuit
        instr_list : list of instructions
            Instruction for the circuit, i.e. [op_name, op_idx, [sites] ]

        Return
        ------
        singvals_cut : list of float
            Singular values cutted, selected through the _trunc_tracking_mode
        """
        singvals_cut = []
        for instr in instr_list:
            sites = instr[2]
            num_sites = len(sites)
            idx = instr[1]
            if instr[0] == "barrier":
                continue

            if num_sites == 1:
                self.emulator.apply_one_site_operator(op_list[idx], *sites)

            elif num_sites == 2:
                singv_cut = self.apply_two_site_gate(op_list[idx], sites[0], sites[1])

                # Avoid errors due to no singv cut
                singv_cut = np.append(singv_cut, 0)
                if self._trunc_tracking_mode == "M":
                    singvals_cut.append(np.max(singv_cut, initial=0.0))
                elif self._trunc_tracking_mode == "C":
                    singvals_cut.append(np.sum(singv_cut**2))

            else:
                raise ValueError("Only one and two-site operations are implemented")
        return singvals_cut

    # pylint: disable-next=too-many-statements, too-many-branches, too-many-locals
    def run_from_qk(self, circuit, operators=None, checkpoints=QCCheckpoints()):
        """
        Run a qiskit quantum circuit on the simulator

        Parameters
        ----------
        circuit : :py:class:`QuantumCircuit`
            qiskit quantum circuit
        operators : TNOperators
            Operators class
        checkpoints : QCCheckpoints
            Checkpoints class

        Returns
        -------
        List[float]
            singular values cutted in the simulation
        Dictionary[TNObservables]
            The dictionary with the observables measured mid circuit
        List[float]
            Memory used in the simulation in bytes
        """
        # data structure of the quantum circuit
        data = circuit.data[checkpoints._initial_line :]
        process = psutil.Process()
        memory = np.zeros(len(data))
        obs_dict = {}
        singvals_cut = []
        for creg in circuit.cregs:
            self.cl_regs[creg.name] = np.zeros(creg.size)

        start_time = time.time()
        barrier_cnt = 0
        # Run over instances
        for idx, instance in enumerate(data):
            operation = instance.operation
            qubits = instance.qubits
            clbits = instance.clbits
            gate_name = operation.name
            num_qubits = len(qubits)
            qubits = [circuit.find_bit(qub).index for qub in qubits]

            # Checking for classical conditions on this gate.
            #
            #  NOTE: Gate.condition will be deprecated in Qiskit 2.0.0
            #  so we need to find an alternative way to make this work.
            #  (https://docs.quantum.ibm.com/api/qiskit/qiskit.circuit.Gate#condition)
            if operation.condition is None:
                apply_gate = True
            else:
                # NOTE: condition should be a tuple (classical_bit, bit_value)
                bit_idx = [clbit.index for clbit in operation.condition[0]]
                bit_value = self.cl_regs[operation.condition[0].name][bit_idx[0]]
                # ^^ possible warning here: we are checking only the first bit_idx

                # Apply the gate only if condition is met:
                apply_gate = bit_value == operation.condition[1]

            # Handling special circuit elements.
            if gate_name == "barrier":
                if self._qc_backend.mpi_settings[barrier_cnt] < 0:
                    self.emulator.reinstall_isometry_serial()
                else:
                    self.emulator.reinstall_isometry_parallel(
                        self._qc_backend.mpi_settings[barrier_cnt]
                    )
                barrier_cnt += 1
                continue
            if gate_name == "measure":
                meas_state, _ = self.apply_projective_operator(*qubits)
                self.cl_regs[clbits[0].register.name][0] = meas_state
                apply_gate = False
            elif gate_name == "reset":
                self.reset(qubits)
                apply_gate = False
            elif gate_name == "MeasureObservables":
                tic = time.time()
                obs = self.meas_observables(operation.observables, operators)
                toc = time.time()
                obs.results_buffer["time"] = tic - start_time
                obs.results_buffer["energy"] = None
                obs.results_buffer["norm"] = self.norm()
                obs.results_buffer["measurement_time"] = toc - tic
                obs_dict[operation.label] = obs
                continue
            if gate_name in ("id", "identity"):
                apply_gate = False
            # possible bug warning:
            # Check that the previous if/elif return either `apply_gate=False`
            # or `continue`. Otherwise, it is expected that `operation` has a
            # method to_matrix(), which is used to apply the gate if apply_gate==True.

            if apply_gate:
                # Grab the operator matrix and move it to the correct device
                gate_mat = operation.to_matrix()
                gate = self.tensor_backend.tensor_cls.from_elem_array(
                    gate_mat, self.tensor_backend.dtype, self.tensor_backend.device
                )
                if num_qubits == 1:
                    self.emulator.apply_one_site_operator(gate, *qubits)
                elif num_qubits == 2:
                    singv_cut = self.apply_two_site_gate(gate, *qubits)
                    singvals_cut += singv_cut
                else:
                    singv_cut = self.apply_multi_site_gate(gate, qubits)
                    singvals_cut += singv_cut

            memory[idx] = process.memory_info().rss
            # Check if you can change settings every n iterations
            self._runtime_checks_updates(idx, self.num_sites, singvals_cut)
            # Save checkpoints if needed
            checkpoints.save_checkpoint(idx, self.emulator)

        return singvals_cut, obs_dict, memory

    # pylint: disable-next=too-many-statements, too-many-branches, too-many-locals
    def run_from_qcirc(self, qcirc, starting_idx=0, checkpoints=QCCheckpoints()):
        """
        Run a simulation starting from a Qcircuit on a portion of the TN state

        Parameters
        ----------
        qcirc : :class:`Qcircuit`
            Quantum circuit
        starting_idx : int, optional
            MPS index that correspond to the index 0 of the Qcircuit. Default to 0.
        checkpoints : QCCheckpoints, optional
            Checkpoints in the simulation

        Returns
        -------
        List[float]
            singular values cutted in the simulation
        Dictionary[TNObservables]
            The dictionary with the observables measured mid circuit
        List[float]
            Memory used in the simulation in bytes
        """
        if not isinstance(qcirc, Qcircuit):
            raise TypeError(f"qcirc must be of type Qcircuit, not {type(qcirc)}")

        process = psutil.Process()
        memory = np.zeros(len(qcirc))
        obs_dict = {}
        singvals_cut = []
        start_time = time.time()
        cnt = -1
        for layer in qcirc:
            for instruction in layer:
                cnt += 1
                if cnt < checkpoints._initial_line:
                    continue
                sites = [ss + starting_idx for ss in instruction[1]]
                operation = instruction[0]

                # Check for classical conditioning
                appy_operation = operation.c_if.is_satisfied(qcirc)
                if appy_operation:
                    # First, check for particular keywords
                    if isinstance(operation, QCObservableStep):
                        operators = (
                            self.tensor_backend.base_tensor_cls.convert_operator_dict(
                                operation.operators,
                                params={},
                                symmetries=[],
                                generators=[],
                                base_tensor_cls=self.tensor_backend.base_tensor_cls,
                                dtype=self.tensor_backend.dtype,
                                device=self.tensor_backend.device,
                            )
                        )
                        tic = time.time()
                        obs = self.meas_observables(
                            operation.observables,
                            operators,
                        )
                        toc = time.time()
                        obs.results_buffer["time"] = tic - start_time
                        obs.results_buffer["norm"] = self.norm()
                        obs.results_buffer["measurement_time"] = toc - tic
                        operation.observables = obs
                        operation.postprocess_obs_indexing()  # Postprocess for qregisters
                        for elem in obs.obs_list:
                            obs.results_buffer.update(obs.obs_list[elem].results_buffer)
                        obs_dict[operation.name] = deepcopy(
                            operation.observables.results_buffer
                        )
                        del obs

                    # Check for particular keywords
                    elif operation.name == "renormalize":
                        self.normalize()
                    elif operation.name == "measure":
                        res = self.emulator.apply_projective_operator(
                            *sites, operation.selected_output
                        )
                        # Update measured value
                        qcirc.modify_cregister(
                            res, operation.cregister, operation.cl_idx
                        )
                    elif operation.name == "add_site":
                        self.emulator.add_site(operation.position)
                    elif operation.name == "remove_site":
                        self.apply_projective_operator(operation.position, remove=True)

                    # Apply gates
                    elif len(sites) == 1:
                        gate = self.tensor_backend.tensor_cls.from_elem_array(
                            operation.operator,
                            self.tensor_backend.dtype,
                            self.tensor_backend.device,
                        )
                        self.site_canonize(*sites, keep_singvals=True)
                        self.apply_one_site_operator(gate, *sites)
                    elif len(sites) == 2:
                        gate = self.tensor_backend.tensor_cls.from_elem_array(
                            operation.operator,
                            self.tensor_backend.dtype,
                            self.tensor_backend.device,
                        )
                        svd_cut = self.apply_two_site_gate(gate, *sites)
                        singvals_cut += svd_cut
                    else:
                        gate = self.tensor_backend.tensor_cls.from_elem_array(
                            operation.operator,
                            self.tensor_backend.dtype,
                            self.tensor_backend.device,
                        )
                        svd_cut = self.apply_multi_site_gate(gate, sites)
                        singvals_cut += svd_cut

                # Check if you can change settings every n iterations
                self._runtime_checks_updates(cnt, self.num_sites, singvals_cut)
                # Save checkpoints if needed
                checkpoints.save_checkpoint(cnt, self.emulator)
                memory[cnt] = process.memory_info().rss

        return singvals_cut, obs_dict, memory

    def _runtime_checks_updates(self, idx, frequency, norm_cut):
        """
        Perform the checks to change the device and the precision if
        idx%frequency is 0.

        Parameters
        ----------
        idx : int
            Index of the current operation of the quantum circuit
        frequency: int
            The checks are done every frequency operations
        norm_cut: float
            The norm cut in the last simulation
        """
        if idx % frequency == 0:
            device = self._qc_backend.resolve_device(
                self.emulator.current_max_bond_dim, self.tensor_backend.device
            )
            precision = self._qc_backend.resolve_precision(
                (1 - np.array(norm_cut)).prod()
            )
            self.emulator.convert(device=device, dtype=precision)


# pylint: disable-next=too-many-statements, too-many-branches, too-many-locals, too-many-arguments
def run_py_simulation(
    circ,
    local_dim=2,
    convergence_parameters=QCConvergenceParameters(),
    operators=QCOperators(),
    observables=TNObservables(),
    initial_state=None,
    backend=QCBackend(),
    checkpoints=QCCheckpoints(),
):
    """
    Transpile the circuit to adapt it to the linear structure of the MPS and run the circuit,
    obtaining in output the measurements.

    Parameters
    ----------
    circ: QuantumCircuit
        qiskit quantum circuit object to simulate
    local_dim: int, optional
        Local dimension of the single degree of freedom. Default is 2, for qubits
    convergence_parameters: :py:class:`QCConvergenceParameters`, optional
        Maximum bond dimension and cut ratio. Default to max_bond_dim=10, cut_ratio=1e-9.
    operators: :py:class:`QCOperators`, optional
        Operator class with the observables operators ALREADY THERE. If None, then it is
        initialized empty. Default to None.
    observables: :py:class:`TNObservables`, optional
        The observables to be measured at the end of the simulation. Default to TNObservables(),
        which contains no observables to measure.
    initial_state : list of ndarray, optional
        Initial state of the simulation. If None, ``|00...0>`` is considered. Default to None.
    backend: :py:class:`QCBackend`, optional
        Backend containing all the information for where to run the simulation
    checkpoints: :py:class:`QCCheckpoints`, optional
        Class to handle checkpoints in the simulation

    Returns
    -------
    result: qmatchatea.SimulationResults
        Results of the simulation, containing the following data:
        - Measures
        - Statevector
        - Computational time
        - Singular values cut
        - Entanglement
        - Measure probabilities
        - MPS state
        - MPS file size
        - Observables measurements
    """
    if isinstance(circ, (QuantumCircuit, Qcircuit)):
        num_qubits = circ.num_qubits
    else:
        raise TypeError(
            "Only qiskit Quantum Circuits and Qcircuit are implemented for"
            + f" simulation, not {type(circ)}"
        )
    start = time.time()
    tensor_backend = _resolve_tensor_backend(
        tensor_module=backend.tensor_module,
        device=backend.resolve_device(1, "cpu"),
        dtype=backend.resolve_precision(1),
    )

    if backend.mpi_approach != "SR" and backend.ansatz == "MPS":
        backend._ansatz = "MPIMPS"

    operators = tensor_backend.base_tensor_cls.convert_operator_dict(
        operators,
        params={},
        symmetries=[],
        generators=[],
        base_tensor_cls=tensor_backend.base_tensor_cls,
        dtype=tensor_backend.dtype,
        device=tensor_backend.device,
    )
    # Check if you selected restart from a checkpoint
    initial_state = checkpoints.restart_from_checkpoint(initial_state)

    # The scalar check is to avoid a warning
    if np.isscalar(initial_state):
        if initial_state is None:
            initial_state = "vacuum"
        simulator = QCEmulator(
            num_qubits,
            convergence_parameters,
            local_dim=local_dim,
            tensor_backend=tensor_backend,
            qc_backend=backend,
            initialize=initial_state.lower(),
        )
    elif isinstance(initial_state, _AbstractTN):
        simulator = QCEmulator.from_emulator(
            initial_state,
            conv_params=convergence_parameters,
            tensor_backend=tensor_backend,
            qc_backend=backend,
        )
    else:
        simulator = QCEmulator.from_tensor_list(
            initial_state,
            conv_params=convergence_parameters,
            tensor_backend=tensor_backend,
            qc_backend=backend,
        )
    if isinstance(circ, QuantumCircuit):
        singvals_cut, obs_dict, memory = simulator.run_from_qk(
            circ, operators, checkpoints=checkpoints
        )
    elif isinstance(circ, Qcircuit):
        singvals_cut, obs_dict, memory = simulator.run_from_qcirc(
            circ, checkpoints=checkpoints
        )
    else:
        # Duplicate from above, but makes linter happy
        raise TypeError(
            "Only qiskit Quantum Circuits and Qcircuit are implemented for pure python"
            + f" simulation, not {type(circ)}"
        )

    end = time.time()

    tic = time.time()
    observables = simulator.meas_observables(observables, operators)
    toc = time.time()

    observables.results_buffer["time"] = end - start
    observables.results_buffer["energy"] = None
    observables.results_buffer["norm"] = simulator.norm()
    observables.results_buffer["measurement_time"] = toc - tic
    observables.results_buffer["memory"] = memory / (1024**3)

    result_dict = observables.results_buffer

    # Observables postprocessing
    postprocess = False
    if simulator.ansatz == "MPIMPS":
        if simulator.rank == 0:
            postprocess = True
    else:
        postprocess = True

    if postprocess:
        for elem in observables.obs_list:
            result_dict.update(observables.obs_list[elem].results_buffer)
            # Special treatment for TNState2file
            if str(elem) == "TNState2File":
                for value in observables.obs_list[elem].name:
                    result_dict["tn_state_path"] = observables.obs_list[
                        elem
                    ].results_buffer[value]

        # Storing the results of measurement happened mid-circuit
        # under their label
        # pylint: disable-next=consider-using-dict-items
        for label in obs_dict:
            obs_values = obs_dict[label]
            tmp = obs_values.results_buffer
            for elem in obs_values.obs_list:
                tmp.update(obs_values.obs_list[elem].results_buffer)
                # Special treatment for TNState2file
                if str(elem) == "TNState2File":
                    for value in obs_values.obs_list[elem].name:
                        tmp["tn_state_path"] = observables.obs_list[
                            elem
                        ].results_buffer[value]

            result_dict[label] = tmp

    results = SimulationResults()
    results.set_results(result_dict, singvals_cut)

    return results


def _resolve_tensor_backend(tensor_module, device, dtype):
    """
    Resolve the string name of the module used for the tensor
    operations.

    Parameters
    ----------
    tensor_module : str
        Name of the module used for the tensor operations

    Returns
    -------
    qtealeaves.tensors._AbstractTensor
    """

    # First fake initialization, to have access to the tensor_cls for
    # the correct dtype
    if tensor_module == "numpy":
        tensor_backend = qtt.TensorBackend()
    elif tensor_module == "torch":
        tensor_backend = qrt.torchapi.default_pytorch_backend()
    elif tensor_module == "tensorflow":
        tensor_backend = qrt.tensorflowapi.default_tensorflow_backend()
    elif tensor_module == "jax":
        tensor_backend = qrt.jaxapi.default_jax_backend()
    else:
        raise ValueError(f"Tensor class with {tensor_module} is not available.")

    # Get the correct dtype
    tmp_tensor = tensor_backend([1, 1])
    dtype = tmp_tensor.dtype_from_char(dtype)

    # Return real tensor backend with correct dtype
    if tensor_module == "numpy":
        return qtt.TensorBackend(device=device, dtype=dtype)
    if tensor_module == "torch":
        return qrt.torchapi.default_pytorch_backend(device=device, dtype=dtype)
    if tensor_module == "tensorflow":
        return qrt.tensorflowapi.default_tensorflow_backend(device=device, dtype=dtype)
    if tensor_module == "jax":
        return qrt.jaxapi.default_jax_backend(device=device, dtype=dtype)

    # Makes linter happy
    raise ValueError(f"Tensor class with {tensor_module} is not available.")



# This code is part of qmatchatea.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""
Function to help in the MPI protocol. Here you insert barriers, which means
canonization of the state each `where_barriers` layers. In particular, it is
important because at each layer in the circuit description is connected a
canonization in the MPS simulation. If the canonization are not often enough
the simulation might not converge.

Functions and classes
---------------------

"""

# pylint: disable=ungrouped-imports, too-many-arguments

import numpy as np
from qiskit import QuantumCircuit

try:
    from qiskit.visualization.utils import _get_layered_instructions
except ImportError:
    # This try except makes the code compatible also with the latest version of
    # qiskit, where the _get_layered_instructions changed path
    from qiskit.visualization.circuit._utils import _get_layered_instructions

__all__ = ["MPISettings", "to_layered_circ"]


class MPISettings:
    """
    Settings for using the library on multiple processes using MPI.
    The default settings use no MPI, i.e. a serial execution of the algorithm.

    Parameters
    ----------
    mpi_approach: str, optional
        Approach for the MPI simulation.
        Available:
        - "SR": serial, the algorithm is run serially on the TN ansatz
        - "CT": cartesian, the MPS is divided between different processes,
                and the algorithm is done in parallel. MPS ONLY.
        Default to "SR".
    isometrization: int | List[int], optional
        How to perform the isometrization.
        - `-1` is a serial isometrization step
        - A `int` is a parallel isometrization step with that number of layers
        - A `List[int]` is a parallel isometrization step where each entry is
          the number of layers for the `i`-th isometrization. If the number of
          isometrization steps is greater than the length of the array, the array
          is repeated.
        Default to `-1`
    num_procs: int, optional
        Number of processes for the MPI simulation. Default to 1.
    mpi_command : List[str], optional
        MPI command that should be called when launching MPI.
        The "-n" for the number of processes is already taken into account.
        Default to ["mpi_exec"].
    where_barriers: int, optional
        This parameter is important only if you want to use MPI parallelization.
        If where_barriers > 0 then the circuit gets ordered in layers and a barrier is applyed
        each where_barriers layers. We recall that a barrier is equivalent to a canonization in
        the MPS simulation. Default to -1.

    """

    def __init__(
        self,
        mpi_approach="SR",
        isometrization=-1,
        num_procs=1,
        mpi_command=None,
        where_barriers=-1,
    ):
        self.num_procs = num_procs
        self.mpi_approach = mpi_approach.upper()
        self.isometrization = isometrization
        if mpi_command is None:
            mpi_command = ["mpiexec"]
        self.mpi_command = list(mpi_command)
        self.where_barriers = where_barriers

    def __getitem__(self, idx):
        """
        Get the settings of the isometrization for the idx-th isometrization
        """
        if np.isscalar(self.isometrization):
            return self.isometrization

        return self.isometrization[idx % len(self.isometrization)]

    def print_isometrization_type(self, idx):
        """
        print which type of isometrization we have at index idx

        Parameters
        ----------
        idx : int
            Index of the isometrization
        """

        iso_type = self[idx]
        if iso_type < 0:
            print(f"At step {idx} serial isometrization")
        else:
            print(f"At step {idx} parallel isometrization with {iso_type} layers")


def _to_layered_circ_qk(qc, where_barriers=1):
    """
    Transform a quantum circuit in another, where the instruction are ordered by layers.
    Apply a barrier every where_barriers layers.

    Parameters
    ----------
    qc: qiskit.QuantumCircuit
        Quantum circuit
    where_barriers: int
        Apply a barrier every where_barriers layers.

    Returns
    -------
    layered_qc: qiskit.QuantumCircuit
            Ordered quantum circuit with barriers
    """
    layered_qc = QuantumCircuit(qc.num_qubits)

    layered_instructions = _get_layered_instructions(qc)[2]
    for ii, layer in enumerate(layered_instructions):
        for instruction in layer:
            layered_qc.append(instruction.op, instruction.qargs, instruction.cargs)
        if ii % where_barriers == 0:
            layered_qc.barrier()

    return layered_qc


def to_layered_circ(circ, where_barriers=1):
    """
    Interface for the transformation of qiskit circuit
    into layered ones, the correct order for the MPI protocols.
    You also insert barriers, i.e. canonization procedures

    Parameters
    ----------
    circ: qiskit.circuit.quantumcircuit.QuantumCircuit
        Quantum circuit instance
    where_barriers: int
        Apply a barrier every where_barriers layers.

    Returns
    -------
    layered_circ: qiskit.circuit.quantumcircuit.QuantumCircuit
        Ordered circuit program with barriers
    """

    if isinstance(circ, QuantumCircuit):
        layered_circ = _to_layered_circ_qk(circ, where_barriers)
    else:
        raise TypeError(
            "Only qiskit quantum circuit is implemented. Your circuit is of type:"
            + str(type(circ))
        )

    return layered_circ



# This code is part of qmatchatea.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""
Module init
"""

__all__ = [
    "utils",
    "preprocessing",
    "interface",
    "par_simulations",
    "circuit",
    "utils",
    "tensor_compiler",
    "__version__",
]
from qmatchatea import circuit, utils
from qmatchatea.interface import *
from qmatchatea.par_simulations import *
from qmatchatea.preprocessing import *
from qmatchatea.tensor_compiler import *
from qmatchatea.utils import (
    QCIO,
    QCBackend,
    QCCheckpoints,
    QCConvergenceParameters,
    QCOperators,
)
from qmatchatea.utils.simulation_results import SimulationResults
from qmatchatea.version import __version__



# This code is part of qmatchatea.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

r"""
Operations that can be applied on the Qcircuit class.
The idea of the module is the following:

- An Operation is whichever kind of operation applied to the circuit
- A ClassicalCondition is a condition on the application of an Operation
  based on the measurement of previous qudits

The available operations in the library at the moment are:

- ``QCMeasureProjective``, apply a projective measurement on the system
- ``QCRenormalize``, renormalize the quantum state to :math:`\sqrt{<\psi|\psi>}=1`
- ``QCAddSite``, add a site to the MPS.
- ``QCRemoveSite``, remove a site from the MPS. It must be in a product state with the rest.
- ``QCObservablesStep``, perform a step of measurement of the observables

The available gates instead are:

- ``QCZpauli``,
- ``QCXpauli``,
- ``QCYpauli``,
- ``QCHadamard``,
- ``QCSgate``,
- ``QCTgate``,
- ``QCSadggate``,
- ``QCTadggate``,
- ``QCPgate``,
- ``QCSwap``,
- ``QCCnot``,
- ``QCCz``
- ``QCCp``

You can always define your own gate through :class:`QCOperation`.

In particular, for the definition of noisy gates, the procedure of the definition of a new
operation is to be preferred over modifying an exsisting one.

Then, the list `[QCOperation, sites]` is called **Instruction**.
"""

import numpy as np

__all__ = [
    "QCOperation",
    "ClassicalCondition",
    "QCMeasureProjective",
    "QCRenormalize",
    "QCAddSite",
    "QCRemoveSite",
    "QCZpauli",
    "QCXpauli",
    "QCYpauli",
    "QCSwap",
    "QCCnot",
    "QCCz",
    "QCHadamard",
    "QCSgate",
    "QCTgate",
    "QCSadggate",
    "QCTadggate",
    "QCPgate",
    "QCCp",
]


class ClassicalCondition:
    """
    Class to handle conditions on classically-controlled
    operations

    Parameters
    ----------
    cregister : str, optional
        Name of the classical register relative to the condition
    value : int, optional
        Value that the register should assume
    idx : int, optional
        If provided, the index of the register to check.
        If not provided, the decimal number connected to the
        binary representation of the register is used.
    """

    def __init__(self, cregister=None, value=None, idx=None):
        self.creg = cregister
        self.value = value
        self.idx = idx

        if cregister is None:
            self._is_initialized = False
        else:
            self._is_initialized = True

    @property
    def is_initialized(self):
        """Initialization property"""
        if self.creg is None:
            self._is_initialized = False
        else:
            self._is_initialized = True

        return self._is_initialized

    def is_satisfied(self, qcirc):
        """
        Check if the classical condition is satisfied

        Parameters
        ----------
        qcirc : :py:class:`QCircuit`
            Quantum circuit class

        Returns
        -------
        bool
            True if the condition is satisfied, False otherwise
        """
        # If the condition is not initialized the condition
        # is automatically satisfied
        if self.is_initialized:
            value = qcirc.inspect_cregister(self.creg, self.idx)
            is_satisfied = value == self.value
        else:
            is_satisfied = True

        return is_satisfied


class QCOperation:
    """
    Operation on the circuit class.

    name: str
        Unique name of the operation
    operator: function
        Python function which returns the array of the operator as np.ndarray
    operator_parameters: list, optional
        List of operator parameters if needed. Default to None.
    conditioned : :py:class:`ClassicalCondition`, optional
        Classical condition to be checked for applying the operation on the
        simulation.
    """

    def __init__(
        self, name, operator, operator_parameters=None, conditioned=ClassicalCondition()
    ):
        self._name = name
        self._operator = operator
        self._operator_parameters = (
            [] if operator_parameters is None else operator_parameters
        )
        if np.isscalar(self._operator_parameters):
            self._operator_parameters = [self._operator_parameters]
        self._condition = conditioned

    def __repr__(self):
        """
        Return the class name as representation
        """
        return self.__class__.__name__

    @property
    def operator(self):
        """
        Returns the operator matrix as np.ndarray
        """
        return self._operator(*self._operator_parameters)

    @property
    def name(self):
        """Name property"""
        return self._name

    @property
    def is_conditioned(self):
        """True if there is a classical condition on the application of the gate"""
        return self._condition.is_initialized

    @property
    def c_if(self):
        """True if there is a classical condition on the application of the gate"""
        return self._condition

    def is_parametric(self):
        """Returns True if the operation is parametrized"""
        is_param = len(self._operator_parameters) == 0
        return is_param

    @property
    def string_rep(self):
        """Return the string representation for printing"""
        str_rep = f"─ {self.name} ─"
        void = " " * len(str_rep)
        string = void + "\n" + str_rep + "\n" + void

        return string


class QCRenormalize(QCOperation):
    """
    Renormalize the state when this operation is encountered

    Parameters
    ----------
    conditioned : :py:class:`ClassicalCondition`, optional
        Classical condition to be checked for applying the operation on the
        simulation.
    """

    def __init__(self, conditioned=ClassicalCondition()):
        QCOperation.__init__(self, "renormalize", None, conditioned=conditioned)


class QCAddSite(QCOperation):
    """
    Add a site to the Qcircuit

    Parameters
    ----------
    position : int
        Position of the link where the site will be added
    conditioned : :py:class:`ClassicalCondition`, optional
        Classical condition to be checked for applying the operation on the
        simulation.
    """

    def __init__(self, position, conditioned=ClassicalCondition()):
        self.position = position
        QCOperation.__init__(self, "add_site", None, conditioned=conditioned)


class QCRemoveSite(QCOperation):
    """
    Remove a site to the Qcircuit

    Parameters
    ----------
    position : int
        Position of the site to remove
    conditioned : :py:class:`ClassicalCondition`, optional
        Classical condition to be checked for applying the operation on the
        simulation.
    """

    def __init__(self, position, conditioned=ClassicalCondition()):
        self.position = position
        QCOperation.__init__(self, "remove_site", None, conditioned=conditioned)


class QCMeasureProjective(QCOperation):
    """
    Apply a projective measurement to the circuit, possibly
    selecting its output

    Parameters
    ----------
    cl_idx : int
        Position of the classical register where the measurement
        will be stored
    selected_output : int, optional
        Output selected a priori. Default to None
    cregister : str, optional
        Classical register where the result of the measurement
        is stored. Default to 'default'
    conditioned : :py:class:`ClassicalCondition`, optional
        Classical condition to be checked for applying the operation on the
        simulation.
    """

    def __init__(
        self,
        cl_idx,
        selected_output=None,
        cregister="default",
        conditioned=ClassicalCondition(),
    ):
        QCOperation.__init__(self, "measure", None, conditioned=conditioned)
        self._selected_output = selected_output
        self.cregister = cregister
        self.cl_idx = cl_idx

    @property
    def selected_output(self):
        """A priori selected output"""
        return self._selected_output


class QCSwap(QCOperation):
    """
    Swap operation between adjacent qubits.
     Its matrix representation is as follows:

    .. math::

        \\mathrm{SWAP} =
        \\begin{pmatrix}
        1 & 0 & 0 & 0 \\\\
        0 & 0 & 1 & 0 \\\\
        0 & 1 & 0 & 0 \\\\
        0 & 0 & 0 & 1 \\\\
        \\end{pmatrix}

    Parameters
    ----------
    conditioned : :py:class:`ClassicalCondition`, optional
        Classical condition to be checked for applying the operation on the
        simulation.
    """

    def __init__(self, conditioned=ClassicalCondition()):
        QCOperation.__init__(self, "swap", None, conditioned=conditioned)

    @property
    def operator(self):
        """
        Overwriting operator with swap
        """
        return np.array([[1, 0, 0, 0], [0, 0, 1, 0], [0, 1, 0, 0], [0, 0, 0, 1]])


class QCCnot(QCOperation):
    """
    Cnot operation between adjacent qubits.
     Its matrix representation is as follows:

    .. math::

        \\mathrm{CNOT} =
        \\begin{pmatrix}
        1 & 0 & 0 & 0 \\\\
        0 & 1 & 0 & 0 \\\\
        0 & 0 & 0 & 1 \\\\
        0 & 0 & 1 & 0 \\\\
        \\end{pmatrix}

    Parameters
    ----------
    conditioned : :py:class:`ClassicalCondition`, optional
        Classical condition to be checked for applying the operation on the
        simulation.
    """

    def __init__(self, swap=True, conditioned=ClassicalCondition()):
        self.swap = swap
        QCOperation.__init__(self, "cx", None, conditioned=conditioned)

    @property
    def operator(self):
        """
        Overwriting operator with cnot
        """
        cnot = np.array([[1, 0, 0, 0], [0, 1, 0, 0], [0, 0, 0, 1], [0, 0, 1, 0]])
        if self.swap:
            swap = QCSwap().operator
            cnot = swap @ cnot @ swap
        return cnot


class QCCz(QCOperation):
    """
    Cz operation between adjacent qubits.
     Its matrix representation is as follows:

    .. math::

        \\mathrm{CZ} =
        \\begin{pmatrix}
        1 & 0 & 0 & 0 \\\\
        0 & 1 & 0 & 0 \\\\
        0 & 0 & 1 & 0 \\\\
        0 & 0 & 0 & -1 \\\\
        \\end{pmatrix}

    Parameters
    ----------
    conditioned : :py:class:`ClassicalCondition`, optional
        Classical condition to be checked for applying the operation on the
        simulation.
    """

    def __init__(self, conditioned=ClassicalCondition()):
        QCOperation.__init__(self, "cz", None, conditioned=conditioned)

    @property
    def operator(self):
        """
        Overwriting operator with cz
        """
        cz_op = np.array([[1, 0, 0, 0], [0, 1, 0, 0], [0, 0, 1, 0], [0, 0, 0, -1]])
        return cz_op

    @property
    def string_rep(self):
        """Return the string representation for printing"""
        str_rep = f"─ {self.name} ─"
        void = " " * len(str_rep)
        string = void + "\n" + str_rep + "\n" + void + "\n"
        string += void + "\n" + str_rep + "\n" + void

        return string


class QCHadamard(QCOperation):
    """
    Hadamard operation.
     Its matrix representation is as follows:

    .. math::

        \\mathrm{H} =\\frac{1}{\\sqrt{2}}
        \\begin{pmatrix}
        1 & 1  \\\\
        1 & -1  \\\\
        \\end{pmatrix}

    Parameters
    ----------
    conditioned : :py:class:`ClassicalCondition`, optional
        Classical condition to be checked for applying the operation on the
        simulation.
    """

    def __init__(self, conditioned=ClassicalCondition()):
        QCOperation.__init__(self, "h", None, conditioned=conditioned)

    @property
    def operator(self):
        """
        Overwriting operator with hadamard
        """
        return np.array([[1, 1], [1, -1]]) / np.sqrt(2)


class QCZpauli(QCOperation):
    """
    Pauli Z operation.
     Its matrix representation is as follows:

    .. math::

        \\mathrm{Z} =
        \\begin{pmatrix}
        1 & 0  \\\\
        0 & -1  \\\\
        \\end{pmatrix}

    Parameters
    ----------
    conditioned : :py:class:`ClassicalCondition`, optional
        Classical condition to be checked for applying the operation on the
        simulation.
    """

    def __init__(self, conditioned=ClassicalCondition()):
        QCOperation.__init__(self, "z", None, conditioned=conditioned)

    @property
    def operator(self):
        """
        Overwriting operator with sigma_z
        """
        return np.array([[1, 0], [0, -1]])


class QCXpauli(QCOperation):
    """
    Pauli X operation.
     Its matrix representation is as follows:

    .. math::

        \\mathrm{X} =
        \\begin{pmatrix}
        0 & 1  \\\\
        1 & 0  \\\\
        \\end{pmatrix}

    Parameters
    ----------
    conditioned : :py:class:`ClassicalCondition`, optional
        Classical condition to be checked for applying the operation on the
        simulation.
    """

    def __init__(self, conditioned=ClassicalCondition()):
        QCOperation.__init__(self, "x", None, conditioned=conditioned)

    @property
    def operator(self):
        """
        Overwriting operator with sigma_x
        """
        return np.array([[0, 1], [1, 0]])


class QCYpauli(QCOperation):
    """
    Pauli Y operation.
     Its matrix representation is as follows:

    .. math::

        \\mathrm{Y} =
        \\begin{pmatrix}
        0 & -i  \\\\
        i & 0  \\\\
        \\end{pmatrix}

    Parameters
    ----------
    conditioned : :py:class:`ClassicalCondition`, optional
        Classical condition to be checked for applying the operation on the
        simulation.
    """

    def __init__(self, conditioned=ClassicalCondition()):
        QCOperation.__init__(self, "y", None, conditioned=conditioned)

    @property
    def operator(self):
        """
        Overwriting operator with sigma_y
        """
        return np.array([[0, -1j], [1j, 0]])


class QCSgate(QCOperation):
    """
    Phase gate with phase :math:`\\pi/2`.
     Its matrix representation is as follows:

    .. math::

        \\mathrm{S} =
        \\begin{pmatrix}
        1 & 0  \\\\
        0 & i  \\\\
        \\end{pmatrix}

    Parameters
    ----------
    conditioned : :py:class:`ClassicalCondition`, optional
        Classical condition to be checked for applying the operation on the
        simulation.
    """

    def __init__(self, conditioned=ClassicalCondition()):
        QCOperation.__init__(self, "s", None, conditioned=conditioned)

    @property
    def operator(self):
        """
        Overwriting operator with phase gate
        """
        return np.array([[1, 0], [0, 1j]])


class QCTgate(QCOperation):
    """
    Phase gate with phase :math:`\\pi/4`.
     Its matrix representation is as follows:

    .. math::

        \\mathrm{T} =
        \\begin{pmatrix}
        1 & 0  \\\\
        0 & \\frac{1}{\\sqrt{2}}(1+i)  \\\\
        \\end{pmatrix}

    Parameters
    ----------
    conditioned : :py:class:`ClassicalCondition`, optional
        Classical condition to be checked for applying the operation on the
        simulation.
    """

    def __init__(self, conditioned=ClassicalCondition()):
        QCOperation.__init__(self, "t", None, conditioned=conditioned)

    @property
    def operator(self):
        """
        Overwriting operator with phase gate
        """
        return np.array([[1, 0], [0, (1 + 1j) / np.sqrt(2)]])


class QCTadggate(QCOperation):
    """
    Phase gate with phase :math:`-\\pi/4`.
     Its matrix representation is as follows:

    .. math::

        \\mathrm{T}^\\dagger =
        \\begin{pmatrix}
        1 & 0  \\\\
        0 & \\frac{1}{\\sqrt{2}}(1-i)  \\\\
        \\end{pmatrix}

    Parameters
    ----------
    conditioned : :py:class:`ClassicalCondition`, optional
        Classical condition to be checked for applying the operation on the
        simulation.
    """

    def __init__(self, conditioned=ClassicalCondition()):
        QCOperation.__init__(self, "tdg", None, conditioned=conditioned)

    @property
    def operator(self):
        """
        Overwriting operator with adjoint t gate
        """
        return np.array([[1, 0], [0, (1 - 1j) / np.sqrt(2)]])


class QCSadggate(QCOperation):
    """
    Adjoint of the phase gate with phase :math:`3\\pi/2`.
     Its matrix representation is as follows:

    .. math::

        \\mathrm{S}^\\dagger =
        \\begin{pmatrix}
        1 & 0  \\\\
        0 & -i  \\\\
        \\end{pmatrix}

    Parameters
    ----------
    conditioned : :py:class:`ClassicalCondition`, optional
        Classical condition to be checked for applying the operation on the
        simulation.
    """

    def __init__(self, conditioned=ClassicalCondition()):
        QCOperation.__init__(self, "sadg", None, conditioned=conditioned)

    @property
    def operator(self):
        """
        Overwriting operator with adjoint s gate
        """
        return np.array([[1, 0], [0, -1j]])


class QCPgate(QCOperation):
    """
    Phase gate with phase :math:`\\theta`.
     Its matrix representation is as follows:

    .. math::

        \\mathrm{P}(\\theta) =
        \\begin{pmatrix}
        1 & 0  \\\\
        0 & e^{i\\theta}  \\\\
        \\end{pmatrix}

    Parameters
    ----------
    conditioned : :py:class:`ClassicalCondition`, optional
        Classical condition to be checked for applying the operation on the
        simulation.
    """

    def __init__(self, theta, conditioned=ClassicalCondition()):
        QCOperation.__init__(self, "p", self.matrix, theta, conditioned=conditioned)

    @staticmethod
    def matrix(theta):
        """
        Overwriting operator with phase gate
        """
        return np.array([[1, 0], [0, np.exp(theta * 1j)]])


class QCCp(QCOperation):
    """
    Controlled Phase gate with phase :math:`\\theta`.
     Its matrix representation is as follows:

    .. math::

        \\mathrm{CP}(\\theta) =
        \\begin{pmatrix}
        1 & 0 & 0 & 0 \\\\
        0 & 1 & 0 & 0 \\\\
        0 & 0 & 1 & 0 \\\\
        0 & 0 & 0 & e^{i\\theta} \\\\
        \\end{pmatrix}

    Parameters
    ----------
    conditioned : :py:class:`ClassicalCondition`, optional
        Classical condition to be checked for applying the operation on the
        simulation.
    """

    def __init__(self, theta, conditioned=ClassicalCondition()):
        QCOperation.__init__(self, "cp", self.matrix, theta, conditioned=conditioned)

    @staticmethod
    def matrix(theta):
        """
        Overwriting operator with controlled phase gate
        """
        cz_op = np.array(
            [[1, 0, 0, 0], [0, 1, 0, 0], [0, 0, 1, 0], [0, 0, 0, np.exp(theta * 1j)]]
        )

        return cz_op



# This code is part of qmatchatea.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""
Class to handle the results of a qmatchatea simulation
"""

# pylint: disable=too-many-instance-attributes
import pickle
from datetime import datetime

import numpy as np
from qtealeaves.emulator import MPS, TTN
from qtealeaves.tensors import TensorBackend

__all__ = ["SimulationResults"]


class SimulationResults:
    """
    Class to store and retrieve the result of a qmatchatea simulation
    """

    def __init__(self, observables=None):
        """
        Initialization.
        If you want to load previous results, just initialize the class
        without parameters.

        Parameters
        ----------
        observables: TNObservables, optional
            observables used in the simulation
        """

        self._datetime = datetime.today().strftime("%Y-%m-%d-%H:%M:%S")

        # Set to None all the variables
        self._initial_state = None
        self._observables = {} if observables is None else observables
        self._singular_values_cut = None
        self._statevector = None

    # ----------------------------
    # Methods to save/load results
    # ----------------------------
    def _store_attr_for_pickle(self):
        """Return dictionary with attributes that cannot be pickled and unset them."""
        storage = {
            "tn_state": self._observables.get("tn_state", None),
            "initial_state": self._initial_state,
            "statevector": self._statevector,
        }

        self._observables["tn_state"] = None
        self._initial_state = None
        self._statevector = None

        return storage

    def _restore_attr_for_pickle(self, storage):
        """Restore attributed removed for pickle from dictionary."""
        # Reset temporary removed attributes
        self._observables["tn_state"] = storage["tn_state"]
        self._initial_state = storage["initial_state"]
        self._statevector = storage["statevector"]

    def save_pickle(self, filename):
        """
        Save class via pickle-module.

        Parameters
        ----------
        filename : str
            path where to save the file

        **Details**

        The following attributes have a special treatment and are not present
        in the copied object.

        * initial state tensor network
        * final state tensor network
        * statevector
        """
        # Temporary remove objects which cannot be pickled which
        # included convergence parameters for lambda function and
        # parameterized variables, the log file as file handle and
        # the MPI communicator
        storage = self._store_attr_for_pickle()

        ext = "pkl"
        if not filename.endswith(ext):
            filename += "." + ext

        with open(filename, "wb") as fh:
            pickle.dump(self, fh)

        self._restore_attr_for_pickle(storage)

    @classmethod
    def read_pickle(cls, filename):
        """
        Load the results of a previous simulation

        Parameters
        ----------
        path: str
            PATH to the file from which we want to load the results

        Returns
        -------
        readme: str
            The text contained in the readme file inside the folder
        """
        ext = "pkl"
        if not filename.endswith(ext):
            raise ValueError(
                f"Filename {filename} not valid, extension should be {ext}."
            )

        with open(filename, "rb") as fh:
            obj = pickle.load(fh)

        if not isinstance(obj, cls):
            raise TypeError(
                f"Loading wrong type as simulation result: {type(obj)} vs {cls}."
            )

        return obj

    def set_results(self, result_dict, singvals_cut):
        """Set the results of a simulation

        Parameters
        ----------
        result_dict: dict
            Dictionary of the attribute to be set
        singvals_cut: array-like
            Array of singular values cut through the simulation
        """
        self._observables = result_dict.copy()
        self._singular_values_cut = np.array(singvals_cut)

        tn_state_path = None
        for key, value in result_dict.items():
            # The only key with the '.' is the MPS state
            if "/" in key:
                tn_state_path = value

        if tn_state_path is not None:
            self._observables["tn_state_path"] = tn_state_path
        self.load_state()

    def load_state(self, tensor_backend=TensorBackend()):
        """
        Loads the state located in `tn_state_path`,
        saving it in `tn_state`

        Parameters
        ----------
        tensor_backend : TensorBackend, optional
            Tensor backend of the state if it is
            saved in a formatted format.
            Default to TensorBackend().
        """
        key = self.observables.get("tn_state_path", None)
        if key is None:
            return

        if key.endswith("pklmps"):
            state = MPS.read_pickle(key)
        elif key.endswith("mps"):
            state = MPS.read(key, tensor_backend=tensor_backend)
        elif key.endswith("pklttn"):
            state = TTN.read_pickle(key)
        elif key.endswith("ttn"):
            state = TTN.read(key, tensor_backend=tensor_backend)
        else:
            raise IOError(f"File extension {key} not supported")

        self._observables["tn_state"] = state

    # -----------------------------
    # Methods to access the results
    # -----------------------------
    @property
    def fidelity(self):
        """
        Return the lower bound for the fidelity of the simulation,
        using the method described in
        If you are interested in the evolution of the fidelity through
        the simulation compute it yourself using `np.cumprod(1-self.singular_values_cut)`.

        Returns
        -------
        float
            fidelity of the final state
        """

        fid = np.prod(1 - np.array(self.singular_values_cut))
        return fid

    @property
    def measures(self):
        """Obtain the measures of the simulation as a dictionary.
        The keys are the measured states, the values the number of occurrencies

        Returns
        -------
        measures: dict
            Measures of the simulation
        """
        return self.observables.get("projective_measurements", None)

    @property
    def statevector(self):
        """Obtain the statevector as a complex numpy array

        Returns
        -------
        statevector: np.array or None
            The statevector of the simulation
        """
        if self._statevector is None:
            if "tn_state" in self.observables.keys():
                tn_state = self.observables["tn_state"]
                if tn_state is None:
                    self.load_state()
                    tn_state = self.observables["tn_state"]
                if tn_state.num_sites < 30:
                    self._statevector = tn_state.to_statevector(
                        qiskit_order=True, max_qubit_equivalent=30
                    )
        return self._statevector.elem

    @property
    def singular_values_cut(self):
        """Obtain the singular values cutted through the simulation, depending on the mode
        chosen. If 'M' for maximum (default), 'C' for cumulated.

        Returns
        -------
        np.ndarray[float]
            Singular values cut during the simulation
        """
        return self._singular_values_cut

    @property
    def computational_time(self):
        """Obtain the computational time of the simulation

        Returns
        -------
        float
            computational time of the simulation
        """
        return self.observables.get("time", None)

    @property
    def entanglement(self):
        """Obtain the bond entanglement entropy measured along each bond of the MPS at
        the end of the simulation

        Returns
        -------
        entanglement: np.array or None
            Bond entanglement entropy
        """
        if "bond_entropy" in self.observables.keys():
            entanglement = self.observables["bond_entropy"]
        elif "bond_entropy0" in self.observables.keys():
            entanglement = self.observables["bond_entropy0"]
        else:
            entanglement = None
        return entanglement

    @property
    def measure_probabilities(self):
        """Return the probability of measuring a given state, which is computed using a
        binary tree by eliminating all the branches with probability under a certain threshold.

        Returns
        -------
        measure_probabilities: Dict[Dict | None]
            probability of measuring a certain state if it is greater than a threshold
        """
        keys = ["unbiased_probability", "even_probability", "greedy_probability"]
        new_keys = ["U", "E", "G"]
        probs = {}
        for key, new_key in zip(keys, new_keys):
            if key in self.observables.keys():
                probs[new_key] = self.observables[key]
            else:
                probs[new_key] = [None]
        return probs

    @property
    def date_time(self):
        """Obtain the starting date and time of the simulation, in the format
        ``Year-month-day-Hour:Minute:Second``

        Returns
        -------
        datetime: string
            The date-time when the simulation started
        """
        return self._datetime

    @property
    def tn_state(self):
        """
        Returns the tensor network class, either TTN or MPS

        Returns
        -------
        _AbstractTN
            The tensor network class
        """
        return self.observables.get("tn_state", None)

    @property
    def tn_state_path(self):
        """
        Returns the tensor list in row-major format.
        The indexing of the single tensor is as follows:

        .. code-block::

            1-o-3
             2|

        Returns
        -------
        mps: list
            list of np.array tensors
        """
        return self.observables.get("tn_state_path", None)

    @property
    def initial_state(self):
        """Returns the initial state of the simulation, as an MPS in row-major format or as
        a string if starting from the Vacuum state

        Returns
        -------
        initial_state: list or str
            list of np.array tensors or Vacuum
        """
        return self._initial_state

    @property
    def observables(self):
        """Returns the expectation values of the observables as a dict with the format
            observable_name : observable_expectation_value

        Returns
        -------
        observables: dict or None
            Expectation values of the observables
        """
        return self._observables



# This code is part of qmatchatea.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

import unittest

import numpy as np
import qtealeaves.observables as obs
from qiskit import QuantumCircuit
from qiskit.quantum_info import Operator

from qmatchatea import run_simulation
from qmatchatea.circuit import Qcircuit, observables
from qmatchatea.circuit.observables import QCObservableStep
from qmatchatea.circuit.operations import QCZpauli
from qmatchatea.py_emulator import QCEmulator
from qmatchatea.utils.tn_utils import QCOperators


class TestQcircuit(unittest.TestCase):
    """Test basic feature of Qcircuit without simulations"""

    def setUp(self):
        """Define 'global' variables"""
        self.num_sites = 4
        # Set seed
        np.random.seed(123)

    def test_addlayer(self):
        """
        Test if adding a layer is working
        """

        qcirc = Qcircuit(self.num_sites)
        qcirc._add_layer()
        self.assertEqual(2, qcirc.num_layers, "Number of layers is 2")

    def test_cregister(self):
        """
        test if the functions connected  to the classical register works
        """
        qcirc = Qcircuit(self.num_sites)
        qcirc.add_cregister("classical", 4)

        self.assertIn("classical", qcirc._cregisters, "classical cregister is created")

        # Fake a measure putting 1 in the 3rd place, with probability 0.5
        qcirc.modify_cregister((1, 0.5), "classical", 2)

        # Check if we can retrive the 1
        self.assertEqual(
            1,
            qcirc.inspect_cregister("classical", 2),
            "Check if we retrieve the 1 in 0100",
        )

        # Check if the relative decimal number is correct
        self.assertEqual(
            4,
            qcirc.inspect_cregister("classical"),
            "Check if we retrieve the decimal" + " number associated with 0100",
        )

    def test_qregister(self):
        """
        Test if the addition/remotion of quantum register works
        This also check for the single site add/remove
        """
        qcirc = Qcircuit(self.num_sites)

        # Add the register
        qcirc.add_qregister("q1", [1, 2, 4])
        self.assertIn("q1", qcirc._qregisters, "q1 qregister is created")

        # Since it has been added in position 1 the first member of the
        # qregister 'q1' should be 1, while the second should be 2+1=3
        # (since it was previously inserted another site)
        self.assertEqual(
            qcirc._qregisters["q1"][0], 1, "First added site is in index 1"
        )
        self.assertEqual(
            qcirc._qregisters["q1"][1], 3, "Second added site is in index 3"
        )

        # Remove the register
        qcirc.remove_qregister("q1")
        self.assertNotIn("q1", qcirc._qregisters, "q1 qregister is removed")

    def test_to_matrix(self):
        """Test the transformation to a matrix in dense and sparse form"""
        num_qubs = 10
        rand_params = np.random.uniform(0, 6, 9)

        # Prepare a qiskit circuit
        qc_qk = QuantumCircuit(num_qubs)
        qc_qk.h(0)
        for ii in range(num_qubs - 1):
            qc_qk.cp(rand_params[ii], ii, ii + 1)
        qc_qk.cx(0, num_qubs - 1)
        qc_qk.cx(1, 4)

        mat = Operator(qc_qk).data

        # And the same as Qcircuit
        qc = Qcircuit(num_qubs)
        qc.h(0)
        for ii in range(num_qubs - 1):
            qc.cp(rand_params[ii], [ii, ii + 1])
        qc.cx([0, num_qubs - 1])
        qc.cx([1, 4])

        qc_mat = qc.to_matrix()
        matrix_is_same = np.isclose(mat, qc_mat).all()

        self.assertTrue(matrix_is_same, "Dense matrix is retrieved correctly")

        qc_mat_sparse = qc.to_matrix(sparse=True).toarray()
        sparse_matrix_is_same = np.isclose(qc_mat_sparse, mat).all()

        self.assertTrue(sparse_matrix_is_same, "Sparse matrix is retrieved correctly")


class TestQcircuitObservables(unittest.TestCase):
    """Test basic feature of Qcircuit without simulations"""

    def setUp(self):
        """Define 'global' variables"""
        self.num_sites = 4
        # Set seed
        np.random.seed(123)
        self.sz = QCZpauli().operator

    def _make_ghz(self, qregs=None):
        qc = Qcircuit(self.num_sites)
        qc.h(0)
        for ii in range(self.num_sites - 1):
            qc.cx([ii, ii + 1])

        if qregs is not None:
            for idx, qreg in enumerate(qregs):
                qc.add_qregister(qreg, np.repeat(idx, idx))

        obsv = obs.TNObservables()
        ops = QCOperators()

        return qc, obsv, ops

    def test_local(self):
        """test local observables"""
        qc, obsv, ops = self._make_ghz()
        sim = QCEmulator(self.num_sites)

        obsv += obs.TNObsLocal("sz", "sz")
        ops["sz"] = self.sz

        qc.measure_observables("step1", obsv, ops, "default")

        qc.add_qregister("new", [0])
        qc.x(0, qreg="new")
        qc.measure_observables("step2", obsv, ops)

        _, results, _ = sim.run_from_qcirc(qc)

        step1_res = results["step1"]["sz"]["default"]
        step2_res = results["step2"]

        expected = np.zeros(4)

        self.assertTrue(
            np.isclose(step1_res, expected).all(),
            "Local measurement on single" + " register correct",
        )

        conds = np.isclose(step2_res["sz"]["default"], expected).all()
        conds = conds and np.isclose(step2_res["sz"]["new"], [-1]).all()
        self.assertTrue(conds, "Local measurement on multiple" + " registers correct")

    def test_projective(self):
        """test projective observables"""
        qc, obsv, ops = self._make_ghz()
        sim = QCEmulator(self.num_sites)

        obsv += obs.TNObsProjective(1024, True)

        qc.measure_observables("step1", obsv, ops, "default")

        qc.add_qregister("new", [0])
        qc.x(0, qreg="new")
        qc.measure_observables("step2", obsv, ops)

        _, results, _ = sim.run_from_qcirc(qc)

        step1_res = results["step1"]["projective_measurements"]["default"]
        step2_res = results["step2"]["projective_measurements"]["new"]

        self.assertIn(
            "1111", step1_res, "Projective measurement on single" + " register correct"
        )

        self.assertEqual(
            step2_res["1"],
            1024,
            "Projective measurement on multiple" + " registers correct",
        )

    def test_probabilities(self):
        """test local probabilities"""
        prob_type = ("U", "E")  # , "G")
        prob_threshold = (0, 0.1)  # , 0.9)
        num_samples = (100, 0)  # , 0) # How does zero samples make sense?
        prob_names = (
            "unbiased_probability",
            "even_probability",
        )  # , "greedy_probability")

        for pn, pb, pt, ns in zip(prob_names, prob_type, prob_threshold, num_samples):
            qc, obsv, ops = self._make_ghz()
            sim = QCEmulator(self.num_sites)

            obsv += obs.TNObsProbabilities(pb, ns, pt)

            qc.measure_observables("step1", obsv, ops, "default")

            qc.add_qregister("new", [0])
            qc.x(0, qreg="new")
            qc.measure_observables("step2", obsv, ops)

            _, results, _ = sim.run_from_qcirc(qc)

            step1_res = results["step1"][pn]["default"]
            step2_res = results["step2"][pn]["new"]

            if pn == "unbiased_probability":
                step2_res["1"] = step2_res["1"][1] - step2_res["1"][0]
                step2_res["0"] = step2_res["0"][1] - step2_res["0"][0]

            step2_res = np.sum(list(step2_res.values()))

            self.assertIn(
                "1111",
                step1_res,
                pb + " probability measurement on single" + " register correct",
            )

            self.assertAlmostEqual(
                step2_res,
                1,
                msg=pb + " probability measurement on multiple" + " registers correct",
            )

    def test_bondentropy(self):
        """test bond entropy observables"""
        qc, obsv, ops = self._make_ghz()
        sim = QCEmulator(self.num_sites)

        obsv += obs.TNObsBondEntropy()

        qc.add_qregister("new", [0])
        qc.x(0, qreg="new")
        qc.cx([0, 0], qreg=["new", "default"])
        qc.measure_observables("step1", obsv, ops)

        _, results, _ = sim.run_from_qcirc(qc)
        bd = results["step1"]["bond_entropy"]

        res = np.array(list(bd.values()))[1:]

        self.assertAlmostEqual(bd[(0, 1)], 0, msg="No entanglement in tensor product")
        self.assertTrue(
            (res == 0.6931471805599454).all(),
            msg="Same GHZ entanglement in all the rest",
        )

    def test_tensor_product(self):
        """test tensor product observables"""
        qc, obsv, ops = self._make_ghz()
        sim = QCEmulator(self.num_sites)
        obsv += obs.TNObsTensorProduct("sz", "sz", 1)
        ops["sz"] = self.sz

        qc.add_qregister("new", [0])
        qc.x(0, qreg="new")
        qc.measure_observables("step1", obsv, ops, qreg="new")

        _, results, _ = sim.run_from_qcirc(qc)

        step2_res = np.real(results["step1"]["sz"])

        self.assertAlmostEqual(
            step2_res, -1, msg="Tensor product operator correctly broadcasted"
        )

    def test_weighted_sum(self):
        """test weighted sum observables"""
        qc, obsv, ops = self._make_ghz()
        sim = QCEmulator(self.num_sites)
        tp = obs.TNObsTensorProduct("sz", ["sz", "id"], [[0], [1]])
        tp += obs.TNObsTensorProduct("sz", ["sz", "id"], [[1], [0]])

        obsv += obs.TNObsWeightedSum("szz", tp, [5, 10])
        ops["sz"] = self.sz
        ops["id"] = np.eye(2)

        qc.add_qregister("new", [0, 0])
        qc.x(0, qreg="new")
        qc.measure_observables("step1", obsv, ops, qreg="new")

        _, results, _ = sim.run_from_qcirc(qc)

        res = np.real(results["step1"]["szz"])

        self.assertAlmostEqual(
            res, 5, msg="Weighted sum operator correctly broadcasted"
        )



# This code is part of qmatchatea.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""
Utility functions to generate common quantum circuits and have
a convenient visualization of the quantum state
"""
from copy import deepcopy

import numpy as np
from qiskit.circuit import Barrier
from qiskit.quantum_info import Statevector

__all__ = ["qiskit_get_statevect", "W_qiskit", "GHZ_qiskit", "QFT_qiskit", "get_index"]


class MeasureObservables(Barrier):
    """
    Class to apply mid-circuit observables measurements with qmatchatea.

    Parameters
    ----------

    label: str
        Name of the observable. It will be used to recover the information
        at the end of the simulation.
    num_qubits: int
        Number of qubits on which the observable is defined. It should be
        the number of qubits in the quantum circuit.
    observables: TNObservables
        The observable class with the information about what you want to
        measure. The input observables will be copied, and it is thus NOT
        modified in place.
    """

    def __init__(self, label, num_qubits, observables):
        self._obs = deepcopy(observables)
        # pylint: disable-next=unexpected-keyword-arg
        super().__init__(num_qubits)
        self._name = "MeasureObservables"
        self._label = label

    @property
    def observables(self):
        """Observables property"""
        return self._obs

    @property
    def label(self):
        """Return the label of the observable"""
        return self._label

    def inverse(self, annotated: bool = False):
        """Special case. Return self."""
        return self


# pylint: disable-next=invalid-name
class qk_transpilation_params:
    """Class to contain the
    transpilation parameters used to map
    a qiskit circuit

    Parameters
    ----------
    linearize: bool, optional
        If True use qiskit transpiler to linearize the circuit. Default to True.
    basis_gates: list, optional
        If not empty decompose using qiskit transpiler into basis gate set
    optimization: integer, optional
        Level of optimization in qiskit transpiler. Default to 1. See
        https://docs.quantum.ibm.com/guides/set-optimization for more details
        about the qiskit transpiler.
    tensor_compiler: bool, optional
        If True, contract all the two-qubit gates before running the experiment.
        Default to False.
    """

    def __init__(
        self, linearize=True, basis_gates=None, optimization=1, tensor_compiler=False
    ):
        self._linearize = linearize
        self._basis_gates = [] if basis_gates is None else basis_gates
        self._optimization = optimization
        self._tensor_compiler = tensor_compiler

    @property
    def linearize(self):
        """linearize property"""
        return self._linearize

    @property
    def basis_gates(self):
        """basis_gates property"""
        return self._basis_gates

    @property
    def optimization(self):
        """optimization property"""
        return self._optimization

    @property
    def tensor_compiler(self):
        """tensor_compiler property"""
        return self._tensor_compiler


def get_index(qc, qubit):
    """
    Get the index of the qubit/bit of a quantum
    circuit.

    Parameters
    ----------
    qc : QuantumCircuit
        qiskit quantum circuit where the qubit is defined
    qubit : Qubit/Bit
        Qubit of which you want to know the index

    Returns
    -------
    int
        Index of the qubit
    """
    return qc.find_bit(qubit)[0]


def qiskit_get_statevect(qc):
    """
    Returns the statevector of the qiskit quantum circuit *qc*

    Parameters
    ----------
    qc: Quantum circuit
        Quantum circuit of which we want the statevector

    Returns
    -------
    st: array_like
        Statevector of the quantum circuit after the application
        of the reverse operation on the qubit's ordering
    """
    qc = deepcopy(qc)

    return Statevector(qc).data


def cphase_swap_qiskit(circuit, control, target, phase):
    """
    Apply to a quantum circuit *circuit* the cphase and swap gate. Acts in place.

    Parameters
    ----------
    circuit: Quantum Circuit
        Qiskit quantum circuit
    control: int
        Index of the control qubit for the controlled phase
    target: int
        Index of the target qubit for the controlled phase
    phase: double
        Phase to apply in the controlled phase in radiants

    Returns
    -------
    None: None
        Acts in place
    """
    circuit.cp(phase, control, target)
    circuit.swap(control, target)


# pylint: disable-next=invalid-name
def F_gate(circuit, kk):
    """
    Apply to a quantum circuit *circuit* the F gate, composed by Ry, CZ, Ry Acts in place.

    Parameters
    ----------
    circuit: Quantum Circuit
        Qiskit quantum circuit
    k: int
        Index of the control qubit for the controlled phase

    Returns
    -------
    None: None
        Acts in place
    """
    nn = circuit.num_qubits
    theta = np.arccos(np.sqrt(1 / (nn - kk)))
    circuit.ry(-theta, kk + 1)
    circuit.cz(kk + 1, kk)
    circuit.ry(theta, kk + 1)


# pylint: disable-next=invalid-name
def QFT_qiskit(circuit, nn):
    """
    Apply the QFT to a qiskit quantum circuit *circuit* in a recursive way

    Parameters
    ----------
        circuit : quantum circuit
            quantum circuit where we want to apply the QFT
        nn       : int
            number of qubits in *circuit*

    Returns
    -------
        None: None
                Acts in place
    """
    # pylint: disable-next=no-else-return
    if nn == 0:
        return circuit
    elif nn == 1:
        circuit.h(0)
        return circuit

    circuit.h(0)
    for ii in range(nn - 1):
        cphase_swap_qiskit(circuit, ii, ii + 1, np.pi * 1 / 2 ** (ii + 1))

    return QFT_qiskit(circuit, nn - 1)


# pylint: disable-next=invalid-name
def GHZ_qiskit(circ):
    """
    Generates a GHZ state in the quantum circuit *circ* composed by n qubits. Acts in place.

    Parameters
    ----------
    circ: Quantum Circuit
        The quantum circuit in the state '00...0' where to build the GHZ

    Returns
    -------
    None: None
        Acts in place.
    """
    nn = circ.num_qubits
    circ.h(0)
    for ii in range(1, nn):
        circ.cx(ii - 1, ii)


# pylint: disable-next=invalid-name
def W_qiskit(circ):
    """
    Generates a W state in the quantum circuit *circ* composed by n qubits. Acts in place.

    Parameters
    ----------
    circ: Quantum Circuit
        The quantum circuit in the state '00...0' where to build the W

    Returns
    -------
    None: None
        Acts in place.
    """
    nn = circ.num_qubits
    circ.x(0)
    for ii in range(nn - 1):
        F_gate(circ, ii)
        circ.barrier()

    for ii in range(nn - 1):
        circ.cx(ii + 1, ii)
        circ.barrier()



# This code is part of qmatchatea.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

import importlib.util
import os
import os.path

import setuptools

# Parse the version file
spec = importlib.util.spec_from_file_location("qmatchatea", "./qmatchatea/version.py")
version_module = importlib.util.module_from_spec(spec)
spec.loader.exec_module(version_module)

# Define requirements
install_requires = [
    "numpy>=1.26.0",
    "scipy>=1.4.1",
    "matplotlib>=3.1.3",
    "qtealeaves>=1.5.8,<1.6.0",
    "qredtea>=0.2.0,<0.3.0",
    "qiskit>=1.0.0,<2.0.0",
    "joblib",
    "psutil",
]

# Get the readme file
if os.path.isfile("README.md"):
    with open("README.md", "r") as fh:
        long_description = fh.read()
else:
    long_description = ""

setuptools.setup(
    name="qmatchatea",
    version=version_module.__version__,
    # Authors alphabetically
    author=", ".join(
        [
            "Marco Ballarin",
            "Francesco Pio Barone",
            "Alberto Coppi",
            "Daniel Jaschke",
            "Guillermo Muñoz Menés",
            "Davide Rattacaso",
            "Nora Reinić",
        ]
    ),
    author_email="quantumtea@lists.infn.it",
    description="Quantum matcha TEA python library for tensor network emulation of quantum circuits.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://baltig.infn.it/quantum_matcha_tea/py_api_quantum_matcha_tea",
    project_urls={},
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ],
    package_dir={
        "qmatchatea": "qmatchatea",
        "qmatchatea.circuit": "qmatchatea/circuit",
        "qmatchatea.utils": "qmatchatea/utils",
    },
    packages=["qmatchatea", "qmatchatea.circuit", "qmatchatea.utils"],
    python_requires=">=3.11",
    install_requires=install_requires,
)



# This code is part of qmatchatea.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

import os
import os.path
import sys
import unittest
from shutil import rmtree

# Try to import cupy
try:
    import cupy as cp
    from cupy_backends.cuda.api.runtime import CUDARuntimeError

    try:
        _ = cp.cuda.Device()
        GPU_AVAILABLE = True
    except CUDARuntimeError:
        GPU_AVAILABLE = False
except ImportError:
    cp = None
    GPU_AVAILABLE = False

from qiskit import QuantumCircuit
from qtealeaves.observables import TNObservables, TNObsTensorProduct, TNObsWeightedSum

from qmatchatea import QCBackend, QCOperators, run_simulation
from qmatchatea.utils import QCIO
from qmatchatea.utils.qk_utils import GHZ_qiskit, W_qiskit

sys.argv = [""]


class TestObservables(unittest.TestCase):
    def setUp(self):
        if not os.path.isdir("TMP_TEST"):
            os.makedirs("TMP_TEST")
        self.qcio = QCIO("TMP_TEST/data/in", "TMP_TEST/data/out")
        self.backends = self.get_python_backends()

    @staticmethod
    def get_python_backends():
        """
        Generate all possible backends for python.
        """
        backends = (
            {
                "precision": "C",
                "device": "cpu",
                "ansatz": "MPS",
            },
            {
                "precision": "Z",
                "device": "cpu",
                "ansatz": "MPS",
            },
            {
                "precision": "C",
                "device": "cpu",
                "ansatz": "TTN",
            },
            {
                "precision": "Z",
                "device": "cpu",
                "ansatz": "TTN",
            },
            {
                "precision": "C",
                "device": "gpu",
                "ansatz": "MPS",
            },
            {
                "precision": "Z",
                "device": "gpu",
                "ansatz": "MPS",
            },
        )

        if GPU_AVAILABLE:
            pass
        else:
            backends = (backends[0], backends[1], backends[2], backends[3])

        backends = [QCBackend(**elem) for elem in backends]

        return backends

    def tearDown(self):
        if os.path.isdir("TMP_TEST"):
            rmtree("TMP_TEST")

    def test_TensorProductOnGHZ(self):
        """
        Test the evaluation of TP observables on GHZ state
        """
        for backend in self.backends:
            approach = backend.identifier

            num_qubits = 16
            qc = QuantumCircuit(num_qubits)
            GHZ_qiskit(qc)

            obs = TNObservables("observables.dat")
            ztp = TNObsTensorProduct(
                "Ztp",
                ["Z" for i in range(num_qubits)],
                [[i] for i in range(num_qubits)],
            )
            xtp = TNObsTensorProduct(
                "Xtp",
                ["X" for i in range(num_qubits)],
                [[i] for i in range(num_qubits)],
            )
            zsingle = TNObsTensorProduct("Zsingle", ["Z"], [[1]])

            obs += xtp
            obs += ztp
            obs += zsingle

            operators = QCOperators()

            res = run_simulation(
                qc,
                io_info=self.qcio,
                observables=obs,
                operators=operators,
                backend=backend,
            )

            obs_res = res.observables
            places = 12 if backend.precision in ("Z", "D") else 5

            self.assertAlmostEqual(
                obs_res["Ztp"],
                1,
                places=places,
                msg=f"Tensor product of ZZ...Z has value 1 on even number of qubits",
            )
            self.assertAlmostEqual(
                obs_res["Xtp"],
                1,
                places=places,
                msg=f"Tensor product of XX...X has value 1",
            )
            self.assertAlmostEqual(
                obs_res["Zsingle"],
                0,
                places=places,
                msg=f"Single Z on second site has value 0",
            )

        return

    def test_TensorProductOnW(self):
        """
        Test the evaluation of TP observables on W state
        """
        for backend in self.backends:
            num_qubits = 16
            qc = QuantumCircuit(num_qubits)
            W_qiskit(qc)

            obs = TNObservables("observables.dat")
            ztp = TNObsTensorProduct(
                "Ztp",
                ["Z" for i in range(num_qubits)],
                [[i] for i in range(num_qubits)],
            )
            xtp = TNObsTensorProduct(
                "Xtp",
                ["X" for i in range(num_qubits)],
                [[i] for i in range(num_qubits)],
            )
            zsingle = TNObsTensorProduct("Zsingle", ["Z"], [[1]])

            obs += xtp
            obs += ztp
            obs += zsingle

            operators = QCOperators()

            res = run_simulation(
                qc,
                io_info=self.qcio,
                observables=obs,
                operators=operators,
                backend=backend,
            )

            obs_res = res.observables

            places = 12 if backend.precision in ("Z", "D") else 5
            self.assertAlmostEqual(
                obs_res["Ztp"],
                -1,
                places=places,
                msg="Tensor product of ZZ...Z has value 0 on even number of qubits",
            )
            self.assertAlmostEqual(
                obs_res["Xtp"],
                0,
                places=places,
                msg="Tensor product of XX...X has value 1",
            )

        return

    def test_WeightedSumOnGHZ(self):
        """
        Test the evaluation of WS observables on GHZ state
        """
        for backend in self.backends:
            num_qubits = 16
            qc = QuantumCircuit(num_qubits)
            GHZ_qiskit(qc)

            obs = TNObservables("observables.dat")
            ztp = TNObsTensorProduct(
                "Ztp",
                ["Z" for i in range(num_qubits - 1)],
                [[i] for i in range(num_qubits - 1)],
            )
            xtp = TNObsTensorProduct(
                "Xtp",
                ["X" for i in range(num_qubits)],
                [[i] for i in range(num_qubits)],
            )

            xx = TNObsWeightedSum("X", xtp, [2], use_itpo=False)
            zz = TNObsWeightedSum("Z", ztp, [3], use_itpo=False)

            obs += xx
            obs += zz

            operators = QCOperators()

            res = run_simulation(
                qc,
                io_info=self.qcio,
                observables=obs,
                operators=operators,
                backend=backend,
            )

            obs_res = res.observables

            places = 12 if backend.precision in ("Z", "D") else 4
            self.assertAlmostEqual(
                obs_res["Z"],
                0,
                places=places,
                msg="Tensor product of ZZ...Z has value 0 on odd number of qubits",
            )
            self.assertAlmostEqual(
                obs_res["X"],
                2,
                places=places,
                msg="Tensor product of XX...X is correct",
            )

        return

    def test_WeightedFromPauli(self):
        """
        Test the evaluation of WS observables on GHZ state
        """
        for backend in self.backends:
            if backend.ansatz == "MPS":
                continue
            num_qubits = 16
            qc = QuantumCircuit(num_qubits)
            GHZ_qiskit(qc)

            obs = TNObservables("observables.dat")
            pauli_dict = {
                "paulis": [
                    {"label": "X" * num_qubits, "coeff": {"real": 1, "imag": 0}},
                    {"label": "Z" * (num_qubits - 1), "coeff": {"real": 1, "imag": 0}},
                ]
            }

            obsw = TNObsWeightedSum.from_pauli_string(
                "test", pauli_dict, use_itpo=False
            )
            obs += obsw

            operators = QCOperators()

            res = run_simulation(
                qc,
                io_info=self.qcio,
                observables=obs,
                operators=operators,
                backend=backend,
            )

            obs_res = res.observables

            places = 12 if backend.precision in ("Z", "D") else 5
            self.assertAlmostEqual(
                obs_res["test"],
                1,
                places=places,
                msg="Tensor product of ZZ...Z + XX...X has value 0 on odd number of qubits",
            )

        return

    def test_WeightedSumOnW(self):
        """
        Test the evaluation of WS observables on W state
        """
        for backend in self.backends:
            num_qubits = 16
            qc = QuantumCircuit(num_qubits)
            W_qiskit(qc)

            obs = TNObservables("observables.dat")
            ztp = TNObsTensorProduct(
                "Ztp",
                ["Z" for i in range(num_qubits)],
                [[i] for i in range(num_qubits)],
            )
            xtp = TNObsTensorProduct(
                "Xtp",
                ["X" for i in range(num_qubits)],
                [[i] for i in range(num_qubits)],
            )

            xx = TNObsWeightedSum("X", xtp, [2])
            zz = TNObsWeightedSum("Z", ztp, [3])
            obs += xx
            obs += zz

            operators = QCOperators()

            res = run_simulation(
                qc,
                io_info=self.qcio,
                observables=obs,
                operators=operators,
                backend=backend,
            )

            obs_res = res.observables

            places = 12 if backend.precision in ("Z", "D") else 5
            self.assertAlmostEqual(
                obs_res["Z"],
                -3,
                places=places,
                msg="Tensor product of ZZ...Z is correct",
            )
            self.assertAlmostEqual(
                obs_res["X"],
                0,
                places=places,
                msg="Tensor product of XX...X is correct",
            )

        return



# This code is part of qmatchatea.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

import sys
import unittest

import numpy as np
from qiskit import QuantumCircuit, transpile, transpiler
from qiskit.circuit.library import QuantumVolume

from qmatchatea import tensor_compiler
from qmatchatea.utils.qk_utils import GHZ_qiskit, W_qiskit, qiskit_get_statevect
from qmatchatea.utils.utils import fidelity

sys.argv = [""]
avail_gates = [
    "x",
    "y",
    "z",
    "h",
    "id",
    "s",
    "sdg",
    "sx",
    "sxdg",
    "t",
    "tdg",
    "swap",
    "dcx",
    "ecr",
    "iswap",
    "ch",
    "cx",
    "cy",
    "cz",
    "p",
    "r",
    "rx",
    "ry",
    "rz",
    "u",
    "u1",
    "u2",
    "rxx",
    "ryy",
    "rzx",
    "rzz",
    "cp",
    "cry",
    "crz",
    "cu",
    "cu1",
]


def qvol_circ(num_qub=3):
    qc = QuantumVolume(num_qub)
    qc = qc.decompose()
    lin_map = transpiler.CouplingMap.from_line(num_qub)
    qc = transpile(qc, coupling_map=lin_map, basis_gates=avail_gates)

    return qc


class TestTensorCompiler(unittest.TestCase):
    def setUp(self):
        np.random.seed(123)

    def test_GHZ(self):
        """
        Check that the GHZ circuit is compiled correctly
        """
        num_qubits = 10
        qc = QuantumCircuit(num_qubits)
        GHZ_qiskit(qc)
        true_state = qiskit_get_statevect(qc)
        compiled_qc = tensor_compiler(qc)
        statevect = qiskit_get_statevect(compiled_qc)
        fid = fidelity(statevect, true_state)
        self.assertAlmostEqual(
            fid, 1, places=12, msg="GHZ state not compiled correctly"
        )

        return

    def test_W(self):
        """
        Check that the W circuit is compiled correctly
        """
        num_qubits = 10
        qc = QuantumCircuit(num_qubits)
        W_qiskit(qc)
        true_state = qiskit_get_statevect(qc)
        compiled_qc = tensor_compiler(qc)
        statevect = qiskit_get_statevect(compiled_qc)
        fid = fidelity(statevect, true_state)
        self.assertAlmostEqual(fid, 1, places=12, msg="W state not compiled correctly")

        return

    def test_QVOLUME(self):
        """
        Check that the QVOLUME circuit is compiled correctly
        """
        num_qubits = 10
        qc = qvol_circ(num_qubits)
        true_state = qiskit_get_statevect(qc)
        compiled_qc = tensor_compiler(qc)
        statevect = qiskit_get_statevect(compiled_qc)
        fid = fidelity(statevect, true_state)
        self.assertAlmostEqual(
            fid, 1, places=12, msg=f"QVOLUME state not compiled correctly"
        )

        return



# This code is part of qmatchatea.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

import io
import os
import subprocess
import unittest
import warnings
from contextlib import redirect_stdout
from pathlib import Path
from shutil import copy, rmtree

import numpy as np
from qtealeaves.tensors.tensor import GPU_AVAILABLE

from examples.advanced.checkpoints import main as checkpoints
from examples.advanced.io_settings import main as io_settings
from examples.advanced.qcemulator import main as qcemulator
from examples.advanced.qudits import main as qudits
from examples.backend.argparse_selector import main as argparse_selector
from examples.backend.device import main as device
from examples.backend.mps_ttn_comparison import main as mps_ttn_comparison
from examples.backend.precisions import main as precisions
from examples.backend.tensor_mods import main as tensor_mods
from examples.circuits.quantum_fourier_transform import (
    main as quantum_fourier_transform,
)
from examples.circuits.random_quantum_circuit import main as random_quantum_circuit
from examples.circuits.teleportation import main as teleportation
from examples.circuits.variational_quantum_eigensolver import (
    main as variational_quantum_eigensolver,
)
from examples.convergence.bond_dimension import main as bond_dimension
from examples.convergence.convergence_analysis import main as convergence_analysis
from examples.convergence.svd_methods import main as svd_methods
from examples.get_started import main as get_started
from examples.observables.bond_entropy import main as bond_entropy
from examples.observables.local import main as local
from examples.observables.mid_circuit_measurement import main as mid_circuit_measurement
from examples.observables.probabilities import main as probabilities
from examples.observables.projective import main as projective
from examples.observables.save_read_results import main as save_read_results
from examples.observables.save_state import main as save_state
from examples.observables.tensor_product import main as tensor_product
from examples.observables.weighted_sum import main as weighted_sum

try:
    import mpi4py

    has_mpi = True
except ImportError:
    has_mpi = False

warnings.filterwarnings("ignore")


class TestExamples(unittest.TestCase):
    """Test that all the examples works correctly"""

    def setUp(self):
        """Define 'global' variables"""
        self.mpi_command = os.environ.get("QMATCHATEA_MPI_COMMAND", "mpiexec")
        self.examples_folder = "examples/"
        if not os.path.isdir(self.examples_folder):
            os.makedirs(self.examples_folder)
        self.num_sites = 4
        # Set seed
        np.random.seed(123)

        self.data_dir = "data"
        if os.path.isdir(self.data_dir):
            rmtree(self.data_dir)

    def tearDown(self):
        if os.path.isdir("TMP_TEST"):
            rmtree("TMP_TEST")
        if os.path.isdir("TMP_MPI"):
            rmtree("TMP_MPI")
        if os.path.isdir("mpi"):
            rmtree("mpi")
        if os.path.isdir(self.data_dir):
            rmtree(self.data_dir)
        return

    def base_mpi(self, filepath, msg):
        """Execute a test with an MPI example."""
        cmd = [self.mpi_command, "-n", "4", "python3", filepath]
        result = subprocess.run(
            cmd,
            capture_output=True,
            text=True,
            env=os.environ,
            cwd=os.getcwd(),
        )

        success = result.returncode == 0
        if not success:
            print("Was running", " ".join(cmd), "in", os.getcwd())
            print("Stdout", result.stdout)
            print("Stderr", result.stderr)
            print("Return code", result.returncode)

        self.assertTrue(success, msg=msg)

    def base_example(self, func):
        """Wrapper for example redirecting output."""
        f = io.StringIO()
        exc = None
        result = None
        with redirect_stdout(f):
            try:
                result = func()
            except Exception as e:
                exc = e

        if exc is not None:
            print("stdout:\n", f.getvalue())
            raise exc

        return result

    def test_check_completeness(self):
        """
        Test whether all the example from the examples folder have been included in the unittest
        """

        this_path = Path(os.path.abspath(__file__))
        examples_path = this_path.parent.parent.absolute() / "examples"
        whitelist = [
            "mpi/data_parallelism_mpi_example.py",  # skipping due to MPI requirement, tested elsewhere
            "mpi/mpi_mps.py",  # skipping due to MPI requirement, tested elsewhere
        ]

        no_missing = True
        missing_list = []

        for file in examples_path.rglob("*.py"):
            file_relative_to_examples = file.relative_to(examples_path)
            file_POSIX_string = str(file_relative_to_examples.as_posix())
            file_as_python_import = ".".join(
                file_relative_to_examples.with_suffix("").parts
            )

            with open(os.path.abspath(__file__), "r") as current_file:
                found = False
                for line in current_file:
                    if line.startswith("from examples." + file_as_python_import):
                        found = True
                        break
                if file_POSIX_string in whitelist:
                    pass
                elif not found:
                    no_missing = False
                    missing_list.append(file_POSIX_string)

        self.assertTrue(
            no_missing,
            "Missing unittest for the following examples files: " + str(missing_list),
        )

    def test_advanced_checkpoints(self):
        """
        Test the checkpoints example
        """
        self.base_example(checkpoints)

    def test_advanced_io_settings(self):
        """
        Test the io_settings example
        """
        self.base_example(io_settings)

    def test_advanced_qcemulator(self):
        """
        Test the qcemulator example
        """
        self.base_example(qcemulator)

    def test_advanced_qudits(self):
        """
        Test the qudits example
        """
        self.base_example(qudits)

    def test_backend_argparse_selector(self):
        """
        Test the argparse_selector example
        """
        self.base_example(argparse_selector)

    @unittest.skipIf(not GPU_AVAILABLE, "GPU is not available")
    def test_backend_device(self):
        """
        Test the device example
        """
        self.base_example(device)

    def test_backend_mps_ttn_comparison(self):
        """
        Test the mps_ttn_comparison example
        """
        self.base_example(mps_ttn_comparison)

    def test_backend_precisions(self):
        """
        Test the precisions example
        """
        self.base_example(precisions)

    def test_backend_tensor_mods(self):
        """
        Test the tensor mods example (numpy, torch, tensorflow, jax).
        Ignore errors if you get the number of backends you expect based
        on your installation, e.g., 2 not equal to 4 if you did install
        only numpy and torch.
        """
        counter = self.base_example(tensor_mods)
        self.assertEqual(
            counter, 4, msg="Failed loading at least one of the four backends."
        )

    def test_circuits_quantum_fourier_transform(self):
        """
        Test the quantum_fourier_transform example
        """
        self.base_example(quantum_fourier_transform)

    def test_circuits_random_quantum_circuit(self):
        """
        Test the random_quantum_circuit example
        """
        self.base_example(random_quantum_circuit)

    def test_circuits_teleportation(self):
        """
        Test the teleportation example
        """
        self.base_example(teleportation)

    def test_circuits_variational_quantum_eigensolver(self):
        """
        Test the variational_quantum_eigensolver example
        """
        self.base_example(variational_quantum_eigensolver)

    def test_convergence_bond_dimension(self):
        """
        Test the bond_dimension example
        """
        self.base_example(bond_dimension)

    def test_convergence_convergence_analysis(self):
        """
        Test the convergence_analysis example
        """
        self.base_example(convergence_analysis)

    def test_convergence_svd_methods(self):
        """
        Test the svd_methods example
        """
        self.base_example(svd_methods)

    # Not sure why the two MPI did have problems, but in the `tests_mpi4py.py`
    # they are tested for now. Ticket is open.

    @unittest.skipIf(not has_mpi, "mpi4py not installed")
    def disabled_mpi_data_parallelism(self):
        """
        Test the mpi_example example
        """
        msg = "mpi data parallelism example simulated parallely fails."
        self.base_mpi(f"{self.examples_folder}mpi/data_parallelism_mpi_example.py", msg)

    @unittest.skipIf(not has_mpi, "mpi4py not installed")
    def disabled_mpi_mpi_mps(self):
        """
        Test the mpi_example example
        """
        msg = "mpi_mpi example simulated parallely on multiple processes fails."
        self.base_mpi(f"{self.examples_folder}mpi/mpi_mps.py", msg)

    def test_observables_bond_entropy(self):
        """
        Test the bond_entropy example
        """
        self.base_example(bond_entropy)

    def test_observables_local(self):
        """
        Test the local example
        """
        self.base_example(local)

    def test_observables_mid_circuit_measurement(self):
        """
        Test the mid_circuit_measurement example
        """
        self.base_example(mid_circuit_measurement)

    def test_observables_probabilities(self):
        """
        Test the probabilities example
        """
        self.base_example(probabilities)

    def test_observables_projective(self):
        """
        Test the projective example
        """
        self.base_example(projective)

    def test_observables_save_read_results(self):
        """
        Test the save_read_results example
        """
        self.base_example(save_read_results)

    def test_observables_save_state(self):
        """
        Test the save_state example
        """
        self.base_example(save_state)

    def test_observables_tensor_product(self):
        """
        Test the tensor_product example
        """
        self.base_example(tensor_product)

    def test_observables_weighted_sum(self):
        """
        Test the weighted_sum example
        """
        self.base_example(weighted_sum)

    def test_get_started(self):
        """
        Test the get_started example
        """
        self.base_example(get_started)
